﻿using System;

namespace GameOfLife
{
  internal class Playfield
  {
    const int OddsForGeneratingCell = 70;
    const int NeighborsCountForCreatingNewCell = 3;

    private readonly int columnCount;
    private readonly int stepCount;
    private readonly Cell[] cells;
    private int step;

    public Playfield(int columnCount, int rowCount, int stepCount)
    {
      this.columnCount = columnCount;
      this.stepCount = stepCount;

      cells = new Cell[this.columnCount * rowCount];
    }

    public void Show()
    {
      Console.Clear();
      Console.WriteLine($"Step {step}/{stepCount}");

      for (int i = 0; i < cells.Length; i++)
      {
        if (i != 0 && i % columnCount == 0)
          Console.WriteLine();

        var value = cells[i] == null ? " " : cells[i].Generation.ToString();

        Console.Write($" {value}");
      }

      Console.WriteLine();
    }

    public void Init()
    {
      step = 1;
      Random random = new Random();

      for (int i = 0; i < cells.Length; i++)
      {
        int odds = 1 + random.Next(0, 100);

        cells[i] = odds <= OddsForGeneratingCell ? null : new Cell();
      }
    }

    public bool Recalculate()
    {
      if (step == stepCount)
        return false;

      step++;
      for (int i = 0; i < cells.Length; i++)
      {
        if (cells[i] is null)
          continue;

        int neighborsCount = GetNeighborsCount(i, 1);

        if (!cells[i].IsAlive(neighborsCount))
          cells[i] = null;
      }

      for (int i = 0; i < cells.Length; i++)
      {
        int neighborsCount = GetNeighborsCount(i, null);

        if (neighborsCount == NeighborsCountForCreatingNewCell)
          cells[i] = new Cell();
      }

      return true;
    }

    // without corners (von Neumann surrounding)
    private int GetNeighborsCount(int cellIndex, int? value)
    {
      var cellRow = cellIndex / columnCount;
      var above = cellIndex - columnCount;
      var under = cellIndex + columnCount;
      var left = cellIndex - 1;
      var right = cellIndex + 1;

      int result = 0;

      if (above >= 0 && IsCellNeighborValid(above, value))
        result++;

      if (under < cells.Length && IsCellNeighborValid(under, value))
        result++;

      if (left >= 0 && left / columnCount == cellRow && IsCellNeighborValid(left, value))
        result++;

      if (right < cells.Length && right / columnCount == cellRow && IsCellNeighborValid(right, value))
        result++;

      return result;
    }

    private bool IsCellNeighborValid(int cellIndex, int? value)
    {
      if (cells[cellIndex] != null && value.HasValue && cells[cellIndex].Generation == value.Value)
        return true;

      return cells[cellIndex] is null && !value.HasValue;
    }
  }
}
