﻿// GameOfLife.h : Include file for standard system include files,
// or project specific include files.

#pragma once

#include <iostream>
#include <vector>
#include <ctime>
#include <string>

class Cell
{
private:
  static const int TooMuchNeighbors = 3;
  static const int TooFewNeighbors = 2;
  char generation;

public:
  Cell()
  {
    generation = 1;
  }

  int GetGeneration()
  {
    return generation;
  }

  bool IsAlive(int neighboursCount)
  {
    if (neighboursCount < TooFewNeighbors)
      return false;

    if (neighboursCount > TooMuchNeighbors)
      return false;

    generation++;

    return true;
  }
};

class Playfield
{
private:
  static const int OddsForGeneratingCell = 70;
  static const int NeighborsCountForCreatingNewCell = 3;

  const int columnCount;
  const int stepCount;
  std::vector<Cell*> cells;
  int step;

public:
  Playfield(int columnCount, int rowCount, int stepCount) : columnCount(columnCount), stepCount(stepCount), cells(columnCount * rowCount)
  {
    step = 1;
  }

  ~Playfield()
  {
    for (int i = 0; i < cells.size(); i++)
    {
      if (cells[i] != nullptr)
        delete cells[i];
    }
  }

  void Show()
  {
    system("cls");
    std::cout << "Step: " << step << "/" << stepCount << std::endl;

    for (int i = 0; i < cells.size(); i++)
    {
      if (i != 0 && i % columnCount == 0)
        std::cout << std::endl;

      auto value = cells[i] == nullptr ? " " : std::to_string(cells[i]->GetGeneration());

      std::cout << " " << value;
    }

    std::cout << std::endl;
  }

  void Init()
  {
    step = 1;
    srand(time(0));

    for (int i = 0; i < cells.size(); i++)
    {
      int odds = 1 + rand() % 100;

      cells[i] = odds <= OddsForGeneratingCell ? nullptr : new Cell();
    }
  }

  bool Recalculate()
  {
    if (step == stepCount)
      return false;

    int cellMarker = 1;

    step++;
    for (int i = 0; i < cells.size(); i++)
    {
      if (cells[i] == nullptr)
        continue;

      int neighborsCount = GetNeighborsCount(i, &cellMarker);

      if (!cells[i]->IsAlive(neighborsCount))
        cells[i] = nullptr;
    }

    for (int i = 0; i < cells.size(); i++)
    {
      int neighborsCount = GetNeighborsCount(i, nullptr);

      if (neighborsCount == NeighborsCountForCreatingNewCell)
        cells[i] = new Cell();
    }

    return true;
  }

private:
  // without corners (von Neumann surrounding)
  int GetNeighborsCount(int cellIndex, int* value)
  {
    auto cellRow = cellIndex / columnCount;
    auto above = cellIndex - columnCount;
    auto under = cellIndex + columnCount;
    auto left = cellIndex - 1;
    auto right = cellIndex + 1;

    int result = 0;

    if (above >= 0 && IsCellNeighborValid(above, value))
      result++;

    if (under < cells.size() && IsCellNeighborValid(under, value))
      result++;

    if (left >= 0 && left / columnCount == cellRow && IsCellNeighborValid(left, value))
      result++;

    if (right < cells.size() && right / columnCount == cellRow && IsCellNeighborValid(right, value))
      result++;

    return result;
  }

  bool IsCellNeighborValid(int cellIndex, int* value)
  {
    if (cells[cellIndex] != nullptr && value != nullptr && cells[cellIndex]->GetGeneration() == *value)
      return true;

    return cells[cellIndex] == nullptr && value == nullptr;
  }
};